#define F_CPU 16000000
#define NUM_LEDS 135 

#define DOT 0
#define SUPERDOT 2
#define PACMAN 3
#define GHOST 4 
#define DEADGHOST 5

#define PACMAN_SPEED 40

#define DIR_STOP 0
#define DIR_RIGHT 1
#define DIR_LEFT -1

#include <util/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include "light_ws2812.h"

struct capsule * pacman;
struct capsule * ghost[4];

uint8_t powercount;
uint8_t initghost;
int8_t lives;
uint8_t restart = 0;
uint8_t manual_mode = 0;

struct capsule {
    uint8_t r;
    uint8_t g;
    uint8_t b;
    uint8_t speed;
    uint8_t speed_i;
    int16_t position;
    uint8_t type; 
    int8_t direction;
}; 

struct node {
    struct capsule *data;
    struct node *next;
};

void delete_next(struct node *t) {
    struct node * tmp = t->next;
    t->next = tmp->next;

    free(tmp->data);
    free(tmp);
}

void insert_after(struct capsule *c, struct node *t) {
    struct node *tmp;
    tmp = (struct node *) malloc(sizeof *tmp);
    if(tmp == NULL) return;
    
    tmp->data = c;
    tmp->next = t->next;
    t->next = tmp;
}

#define BIT_S(var,b) ((var&(1<<b))?1:0)

uint8_t easy_random() {
    static uint16_t m = 0xaa;

    uint8_t x;
    for (x = 0; x < 8; x++){
        m = (m<<1) ^ BIT_S(m,1) ^ BIT_S(m,8) ^ BIT_S(m,9) ^ BIT_S(m,13) ^ BIT_S(m,15);
    }
    return (uint8_t) m;
}


struct CRGB { uint8_t g; uint8_t r; uint8_t b; };
struct CRGB led[NUM_LEDS];

void winner (uint8_t type) {

    int16_t i;
    if (type == GHOST) {
        
        struct CRGB color = { 0, 0, 8 };
        for (i=NUM_LEDS-1;i>=0;i--) {
            led[i] = color;
            ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS);
            _delay_ms(10);
        } 
        ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS);
        
    } else if (type == PACMAN) {
        struct CRGB color = { 8, 9, 0 };
        
        for (i=0;i<NUM_LEDS;i++) {
            led[i] = color;
            ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS);
            _delay_ms(10);
        }
        ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS); 
    }

    _delay_ms(2000);
    _delay_ms(2000);
    restart = 1; 
}

void blink_pixel(uint8_t pos) {

    struct CRGB blank = { 0, 0, 0 };
    struct CRGB color = led[pos];

    ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS);
    _delay_ms(250);

    led[pos] = blank;
    ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS);
    _delay_ms(250);
    led[pos] = color;
    ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS);
    _delay_ms(250);
    led[pos] = blank;
    ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS);
    _delay_ms(250);
}

void kill_ghost(struct capsule * ghost) {
    blink_pixel(ghost->position);
    ghost->r = 3;
    ghost->g = 3;
    ghost->b = 10;
    ghost->direction = DIR_RIGHT;
    ghost->speed = ghost->speed_i = PACMAN_SPEED / 4;

    ghost->type = DEADGHOST;
}

void kill_pacman() {
    blink_pixel(pacman->position);
    pacman->position = -3;
    pacman->direction = DIR_RIGHT;
    pacman->speed = PACMAN_SPEED;

    struct CRGB blank = { 0, 0, 0 };

    if (lives > 0) {
        uint8_t i;
        for (i=0;i<lives;i++) {
            led[i].r = pacman->r;
            led[i].g = pacman->g;
            led[i].b = pacman->b;
        }
        ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS);
        _delay_ms(500);
        blink_pixel(lives-1);
        led[lives-1] = blank;
        ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS);
        _delay_ms(1000);
    }
    lives--;
}


int8_t check_collision(struct capsule * c, int8_t d) {
    if(pacman->position < 0 || c->position > NUM_LEDS) return 0;     
   
    if (c->type == PACMAN && c->position >= 0) {
        int i;
        for (i=0;i<4;i++) {
            if (((c->position + d + NUM_LEDS) % NUM_LEDS) == ghost[i]->position && ghost[i]->type == GHOST) {
                // pacman pokes ghost
                if (powercount > 0) {
                    kill_ghost(ghost[i]);
                } else {
                    kill_pacman();
                }
                return 1;
            }
        }
    } else if (c->type == GHOST) {
        if (((c->position + d + NUM_LEDS) % NUM_LEDS) == pacman->position) {
            // ghost collides pacman
            if (powercount > 0) {
                kill_ghost(c);
            } else {
                kill_pacman();
            }
            return 1;
        }
    }
    return 0;
}


void copy_to_buffer(struct node *head) {

    struct CRGB blank = { 0, 0, 0 };
    
    uint16_t i;
    for (i = 0; i < NUM_LEDS; i++) {
        led[i] = blank;
    }

    struct node * n = head->next;
    while (n != NULL) {
        if (n->data->position >= 0 && n->data->position < NUM_LEDS) { 
            if (n->data->type == GHOST && (powercount > 8 || (powercount > 0 && n->data->position % 2 && n->data->speed_i > 0))) {
                led[n->data->position].r += 0;
                led[n->data->position].g += 0;
                led[n->data->position].b += 30;
            } else {
                led[n->data->position].r += n->data->r;
                led[n->data->position].g += n->data->g;
                led[n->data->position].b += n->data->b;
            }
        }
        n = n->next;
    }
}

void do_move(struct capsule *c) {
    c->speed_i = c->speed;
    if(check_collision(c, c->direction)) return; // do not move if there would be a collision
    
    switch(c->direction) {
        case DIR_LEFT:
            if(c->position > 0) {
                c->position--;
            } else {
                c->position = NUM_LEDS - 1;
            }
            break;
        
        case DIR_RIGHT:
            if(c->position + 1 < NUM_LEDS) {
                c->position++;
            } else if(c->type == DEADGHOST) {
                c->position = NUM_LEDS + 1;
                c->direction = DIR_STOP;
            } else {
                c->position = 0;
            }
            break;
            
        default: break; 
    }   
}

void animate_ghosts() {
    uint8_t i;
    for(i=0; i<4; i++) {
        if(ghost[i]->type == DEADGHOST && ghost[i]->direction == DIR_STOP) continue;
        
        if(ghost[i]->speed_i == 0) {
            do_move(ghost[i]);
            
            if(initghost == 0 && ghost[i]->type != DEADGHOST && easy_random() < 7) {
                ghost[i]->direction *= -1;
            }
        } else {
            ghost[i]->speed_i--;
        }
    }
}

void animate_pacman(struct node *head) {
    struct node *n;
    
    if(pacman->speed_i == 0) {
        do_move(pacman);
        
        // eat dot?
        n = head;
        while(n->next != NULL) {
            if((n->next->data->type == DOT || n->next->data->type == SUPERDOT) && pacman->position == n->next->data->position) {
                if(n->next->data->type == SUPERDOT) {
                    pacman->speed = pacman->speed_i = PACMAN_SPEED / 1.5;
                    powercount = 30;
                }
                
                // eat the dot
                delete_next(n);
                break; // loop can be aborted now, because there is max. 1 dot on this position
            }
            
            n = n->next;
        }

        // pacman brain 
        uint8_t dist = NUM_LEDS;
        uint8_t i;
        for(i=0; i<4; i++) {
            if (ghost[i]->type == GHOST && ghost[i]->direction != pacman->direction && powercount == 0) {
                uint8_t j;
                for(j=0; j<NUM_LEDS; j++) {
                    if ((pacman->position + (j * pacman->direction) + NUM_LEDS) % NUM_LEDS == ghost[i]->position) {
                        dist = j < dist ? j : dist;
                        break;
                    }
                }
            }
        }

        if (!manual_mode) {
            if (dist < 5 && easy_random() % 3 == 1) {
                pacman->direction *= -1;
            } else if (dist < 10 && easy_random() % 6 == 1) {
                pacman->direction *= -1;
            }
            if (easy_random() <=1) pacman->direction *= -1;
        }

        // decrease powercount
        if(powercount > 0) {
            powercount--;
            if(powercount == 0) {
                pacman->speed = pacman->speed_i = PACMAN_SPEED;
            }
        }
            
        
        if(initghost > 0) initghost--;
    } else {
        pacman->speed_i--;
    }
}

void init_pacman(struct node * head) {
    pacman = malloc(sizeof (struct capsule));
    if (pacman != NULL) {
        pacman->r = 35;
        pacman->g = 32;
        pacman->b = 4;
        pacman->speed = PACMAN_SPEED;
        pacman->speed_i = PACMAN_SPEED;
        pacman->position = -1;
        pacman->type = PACMAN;
        pacman->direction = DIR_RIGHT;

        insert_after(pacman, head);
    }
}

void init_ghosts(struct node * head) {
    
    struct capsule * inky = malloc(sizeof (struct capsule));
    inky->r = 1; // 5
    inky->g = 12; // 13
    inky->b = 12; // 20
    inky->speed = PACMAN_SPEED;
    inky->speed_i = PACMAN_SPEED;
    inky->position = NUM_LEDS+10;
    inky->type = GHOST;
    inky->direction = DIR_LEFT;

    ghost[0] = inky;
    insert_after(inky, head);
    

    struct capsule * blinky = malloc(sizeof (struct capsule));
    blinky->r = 14;
    blinky->g = 2;
    blinky->b = 2;
    blinky->speed = PACMAN_SPEED;
    blinky->speed_i = PACMAN_SPEED;
    blinky->position = NUM_LEDS+12;
    blinky->type = GHOST;
    blinky->direction = DIR_LEFT;

    ghost[1] = blinky;
    insert_after(blinky, head);


    struct capsule * pinky = malloc(sizeof (struct capsule));
    pinky->r = 24;
    pinky->g = 12;
    pinky->b = 14;
    pinky->speed = PACMAN_SPEED;
    pinky->speed_i = PACMAN_SPEED;
    pinky->position = NUM_LEDS+14;
    pinky->type = GHOST;
    pinky->direction = DIR_LEFT;

    ghost[2] = pinky;
    insert_after(pinky, head);


    struct capsule * clyde = malloc(sizeof (struct capsule));
    clyde->r = 20;
    clyde->g = 10;
    clyde->b = 0;
    clyde->speed = PACMAN_SPEED;
    clyde->speed_i = PACMAN_SPEED;
    clyde->position = NUM_LEDS+16;
    clyde->type = GHOST;
    clyde->direction = DIR_LEFT;

    ghost[3] = clyde;
    insert_after(clyde, head);
}

void init_board(struct node * head) {
    int i;
    for (i = 0; i < NUM_LEDS; i+=2) {
        struct capsule * a = malloc(sizeof (struct capsule));
        if (a != NULL) {
            if (i > 0 && i % 50 == 0) {
                a->r = 8;
                a->g = 8;
                a->b = 12;
                a->type = SUPERDOT;
            } else { 
                a->r = 1;
                a->g = 1;
                a->b = 1;
                a->type = DOT;
            }
            a->position = i;

            insert_after(a, head);
        }

        copy_to_buffer(head);
        ws2812_sendarray((uint8_t *)&led[0], 3 * (i+1));
        _delay_ms(20);
    }
}


void init(struct node * head) {

    powercount = 0;
    initghost = 15;
    lives = 3;

    while(head->next != NULL) {
        delete_next(head);
    }

    init_board(head);
    init_pacman(head);
    init_ghosts(head);
}



int main(void) {
    DDRB |= _BV(PB0);
    DDRD |= _BV(PD5); // RTS-simulation on TxD-Pin
    DDRF |= _BV(PD4); // there is no PF4 defined :(

    PORTD |= (1<<PD2); // activate internal pullupPullup
    // button on RxD / INT2 / PD2

    struct node * head = malloc(sizeof *head);
    head->next = NULL;

    RESTART:
    restart = 0; 
    manual_mode = 0;
    init(head);

    int debounce = 0;

    while(1) {
        if ((PIND & (1 << PD2)) > 0) {
            if (!debounce) {
                manual_mode = 1;
                pacman->direction *= -1;
                debounce = 10;
            }
        } else {
            if (debounce > 0) debounce--; 
        }

        animate_ghosts();
        copy_to_buffer(head);

        animate_pacman(head);
        copy_to_buffer(head);
        
        ws2812_sendarray((uint8_t *)&led[0], 3 * NUM_LEDS);


        int8_t dots_left = 0;
        struct node *n = head;
        while (n->next != NULL) {
            if (n->next->data->type == DOT || n->next->data->type == SUPERDOT) {
                dots_left=1;
                break;
            }
            n = n->next;
        }

        if (!dots_left) {
            winner(PACMAN);
        }
        if (lives < 0) {
            winner(GHOST);
        }

        if (restart > 0) {
            goto RESTART;
        }
    }
}
